import networkx
from node2vec import Node2Vec
import numpy as np
import time


class Embedder:
    def __init__(self, adjacency_matrix=None, graph=None):
        """
        Embed graph to an n-dimensional embedding space
        A graph or an adjacency matrix should be provided. If graph is provided, adjacency matrix will be ignored.

        :param adjacency_matrix: scipy sparse CSR
        :param graph: networkx graph
        """
        if graph:
            self.graph = graph
        else:
            self.graph = networkx.from_scipy_sparse_matrix(
                abs(adjacency_matrix))
        if graph is None and adjacency_matrix is None:
            raise Exception('Graph or AdjacencyMatrix should be provided.')

    def node2vec(self, dimensions, walk_length, num_walks, workers=1, p=1, q=1):
        """
        Generates embedding space from the given graph/adjacency_matrix in a way that similar nodes in the graph are close to each other in the embedding space.
        For more information, visit https://snap.stanford.edu/node2vec/

        :param dimensions: Embedding space dimensions
        :param walk_length: Length of random-walk
        :param num_walks: Number of random-walks performed
        :param workers: Number of workers available for parallel computation, default: 1
        :param p: Return parameter, default: 1
        :param q: In-Out parameter, default: 1

        :return: embedding space as a numpy array of shape (number_of_nodes, dimensions)
        """
        init_time = time.time()
        model = Node2Vec(
            self.graph,
            dimensions=dimensions,
            walk_length=walk_length,
            num_walks=num_walks,
            workers=workers,
            p=p,
            q=q
        )
        # Fit node2vec model
        model = model.fit()

        # Extract embedding vectors for each node
        embedding_vectors = np.zeros(
            (self.graph.number_of_nodes(), dimensions))
        for i in range(self.graph.number_of_nodes()):
            embedding_vectors[i] = model.wv.get_vector(str(i))

        print(
            f'Elapsed time (node2vec embedding): {(time.time() - init_time):.2f}s')

        return embedding_vectors



from .embedding import Embedder
from .clustering import Cluster
from scipy.sparse import lil_matrix


class Coarsener:
    def __init__(self, adjacency_matrix):
        """
        Generate a coarser grid from a given fine adjacency matrix

        :param adjacency_matrix: scipy sparse CRS
        """
        self.adjacency_matrix = adjacency_matrix

    def apply(
        self,
        dimensions,
        walk_length,
        num_walks,
        p,
        q,
        number_of_clusters=None,
        clustering_method='kmeans',
        workers=1
    ):
        """
        Generates coarse level adjacency matrix using GL-Coarsener method, as proposed in https://arxiv.org/abs/2011.09994

        :param dimensions: `Embedding space dimensions`
        :param walk_length: Length of random-walk
        :param num_walks: Number of random-walks performed
        :param p: Return parameter
        :param q: In-Out parameter
        :param number_of_clusters: Size of the coarse matrix, default: adjacency_matrix.shape[0] // 5
        :param clustering_method: Clustering method to be used; could be 'kmeans' of 'minibatch_kmeans, default: 'kmeans'
        :param workers: Number of workers available for parallel computation, default: 1

        :rtype: scipy sparse CRS of size (number_of_clusters, number_of_clusters)
        """
        # Initial value for number of clusters
        if number_of_clusters is None:
            number_of_clusters = self.adjacency_matrix.shape[0] // 5

        # Step 1: Generate embeddin vectors from the given adjacency matrix (generate embedding space)
        embedder = Embedder(adjacency_matrix=self.adjacency_matrix)
        embedding_vectors = embedder.node2vec(
            dimensions=dimensions,
            walk_length=walk_length,
            num_walks=num_walks,
            workers=workers,
            p=p,
            q=q
        )

        # Step 2: Feed the embedding vectors to clustering module and generate clusters of nodes
        cluster = Cluster(embedding_vectors=embedding_vectors)
        if clustering_method == 'kmeans':
            groups = cluster.kmeans(number_of_clusters=number_of_clusters)
        elif clustering_method == 'minibatch_kmeans':
            groups = cluster.minibatch_kmeans(
                number_of_clusters=number_of_clusters)

        # Step 3: Compute prolongation and restriction operators
        # Manipulating LIL sparse matrix is more efficient than CSR sparse matrix
        p = lil_matrix((self.adjacency_matrix.shape[0], number_of_clusters))
        for i in range(self.adjacency_matrix.shape[0]):
            for j in range(len(groups)):
                p[i, j] = 1 if i in groups[j] else 0
        # Performing matrix operations on CSR sparse matrix is more efficient than LIL sparse matrix
        p = p.tocsr()
        r = p.transpose()

        # Step 4: Compute coarse adjacency matrix
        coarse_adjacency_matrix = r.dot(self.adjacency_matrix).dot(p)

        return coarse_adjacency_matrix



from sklearn.cluster import KMeans, MiniBatchKMeans
import time


class Cluster:
    def __init__(self, embedding_vectors):
        """
        An embeddding space (embedding vectors) should be provided
        """
        self.embedding_vectors = embedding_vectors

    def kmeans(self, number_of_clusters):
        """
        Cluster nodes in the embedding space using KMeans clustering algorithm

        :param number_of_clusters: Number of clusters to be generated
        :return: Generated clusters
        """
        init_time = time.time()
        clusters = KMeans(
            init='random',
            n_clusters=number_of_clusters,
            verbose=1
        ).fit(self.embedding_vectors)

        # Generate a list of clustered nodes
        groups = [[] for _ in range(max(clusters.labels_) + 1)]
        for node_index, group_index in enumerate(clusters.labels_):
            groups[group_index].append(node_index)

        print(
            f'Elapsed time (KMeans clustering): {(time.time() - init_time):.2f}s')
        return groups

    def minibatch_kmeans(self, number_of_clusters):
        """
        Cluster nodes in the embedding space using Mini-Batch KMeans clustering algorithm

        number_of_clusters: Number of clusters to be generated
        :return: Generated clusters
        """
        init_time = time.time()
        clusters = MiniBatchKMeans(
            init='random',
            init_size=number_of_clusters * 5,
            n_clusters=number_of_clusters,
            verbose=1,
            batch_size=(number_of_clusters // 20) + 1,
            reassignment_ratio=1,
            max_iter=500,
            n_init=2,
            max_no_improvement=50
        ).fit(self.embedding_vectors)

        # Generate a list of clustered nodes
        groups = [[] for _ in range(max(clusters.labels_) + 1)]
        for node_index, group_index in enumerate(clusters.labels_):
            groups[group_index].append(node_index)

        print(
            f'Elapsed time (Mini-Batch KMeans clustering): {(time.time() - init_time):.2f}s')
        return groups



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="gl-coarsener",  # Replace with your own username
    version="1.0.1",
    author="Reza Namazi",
    author_email="rezanmz@ymail.com",
    description="GL-Coarsener, a graph learning based coarsening method",
    keywords=['multigrid', 'gl-coarsener', 'graph', 'machine learning',
              ' graph embedding', 'representation learning'],
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/rezanmz/GL-Coarsener",
    packages=setuptools.find_packages(),
    install_requires=['node2vec', 'scipy',
                      'numpy', 'networkx', 'scikit-learn'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering :: Mathematics",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Intended Audience :: Science/Research",
    ],
    python_requires='>=3.6',
)
